/**
 * Fichier pricipal du générateur de code Luge
 */


'use strict';

goog.provide('Blockly.Luge');

goog.require('Blockly.Generator');

function guid() {
  function s4() {
    return Math.floor((1 + Math.random()) * 0x10000)
      .toString(16)
      .substring(1);
  }
  return s4() + s4() + '-' + s4() + '-' + s4() + '-' +
    s4() + '-' + s4() + s4() + s4();
}

function indent(str, numOfIndents, opt_spacesPerIndent) {
  var lines = str.split('\n');
  var indent_lines = [];
  
  for (var line in lines) {
	  indent_lines.push(new Array(opt_spacesPerIndent + 1 || 0).join(' ') + lines[line]);
  }
  
  return indent_lines.join('\n');
}


/**
 * Luge code generator.
 * @type {!Blockly.Generator}
 */
Blockly.Luge = new Blockly.Generator('Luge');

Blockly.Luge.FileTAG = guid();

/**
 * List of illegal variable names.
 * This is not intended to be a security feature.  Blockly is 100% client-side,
 * so bypassing this list is trivial.  This is intended to prevent users from
 * accidentally clobbering a built-in object or function.
 * @private
 */
Blockly.Luge.addReservedWords(
    // import keyword
    // print(','.join(sorted(keyword.kwlist)))
    // https://docs.python.org/3/reference/lexical_analysis.html#keywords
    // https://docs.python.org/2/reference/lexical_analysis.html#keywords
    'Declare,Inclure,Debut,Fin,Appel,'+
	'V:,O:,C:,N:,G:,I:,F:,T:,E:,P:,'+
	'Chaine,Numérique,Booléen'
);

/**
 * Order of operation ENUMs.
 * http://docs.python.org/reference/expressions.html#summary
 */
Blockly.Luge.ORDER_ATOMIC = 0;            // 0 "" ...
Blockly.Luge.ORDER_COLLECTION = 1;        // tuples, lists, dictionaries
Blockly.Luge.ORDER_STRING_CONVERSION = 1; // `expression...`
Blockly.Luge.ORDER_MEMBER = 2.1;          // . []
Blockly.Luge.ORDER_FUNCTION_CALL = 2.2;   // ()
Blockly.Luge.ORDER_EXPONENTIATION = 3;    // **
Blockly.Luge.ORDER_UNARY_SIGN = 4;        // + -
Blockly.Luge.ORDER_BITWISE_NOT = 4;       // ~
Blockly.Luge.ORDER_MULTIPLICATIVE = 5;    // * / // %
Blockly.Luge.ORDER_ADDITIVE = 6;          // + -
Blockly.Luge.ORDER_BITWISE_SHIFT = 7;     // << >>
Blockly.Luge.ORDER_BITWISE_AND = 8;       // &
Blockly.Luge.ORDER_BITWISE_XOR = 9;       // ^
Blockly.Luge.ORDER_BITWISE_OR = 10;       // |
Blockly.Luge.ORDER_RELATIONAL = 11;       // in, not in, is, is not,
                                            //     <, <=, >, >=, <>, !=, ==
Blockly.Luge.ORDER_LOGICAL_NOT = 12;      // not
Blockly.Luge.ORDER_LOGICAL_AND = 13;      // and
Blockly.Luge.ORDER_LOGICAL_OR = 14;       // or
Blockly.Luge.ORDER_CONDITIONAL = 15;      // if else
Blockly.Luge.ORDER_LAMBDA = 16;           // lambda
Blockly.Luge.ORDER_NONE = 99;             // (...)

/**
 * List of outer-inner pairings that do NOT require parentheses.
 * @type {!Array.<!Array.<number>>}
 */
Blockly.Luge.ORDER_OVERRIDES = [
  // (foo()).bar -> foo().bar
  // (foo())[0] -> foo()[0]
  [Blockly.Luge.ORDER_FUNCTION_CALL, Blockly.Luge.ORDER_MEMBER],
  // (foo())() -> foo()()
  [Blockly.Luge.ORDER_FUNCTION_CALL, Blockly.Luge.ORDER_FUNCTION_CALL],
  // (foo.bar).baz -> foo.bar.baz
  // (foo.bar)[0] -> foo.bar[0]
  // (foo[0]).bar -> foo[0].bar
  // (foo[0])[1] -> foo[0][1]
  [Blockly.Luge.ORDER_MEMBER, Blockly.Luge.ORDER_MEMBER],
  // (foo.bar)() -> foo.bar()
  // (foo[0])() -> foo[0]()
  [Blockly.Luge.ORDER_MEMBER, Blockly.Luge.ORDER_FUNCTION_CALL],

  // not (not foo) -> not not foo
  [Blockly.Luge.ORDER_LOGICAL_NOT, Blockly.Luge.ORDER_LOGICAL_NOT],
  // a and (b and c) -> a and b and c
  [Blockly.Luge.ORDER_LOGICAL_AND, Blockly.Luge.ORDER_LOGICAL_AND],
  // a or (b or c) -> a or b or c
  [Blockly.Luge.ORDER_LOGICAL_OR, Blockly.Luge.ORDER_LOGICAL_OR]
];

/**
 * Initialise the database of variable names.
 * @param {!Blockly.Workspace} workspace Workspace to generate code from.
 */
Blockly.Luge.init = function(workspace) {
  /**
   * Empty loops or conditionals are not allowed in Python.
   */
  Blockly.Luge.PASS = this.INDENT + 'pass\n';
  // Create a dictionary of definitions to be printed before the code.
  Blockly.Luge.definitions_ = Object.create(null);
  // Create a dictionary mapping desired function names in definitions_
  // to actual function names (to avoid collisions with user functions).
  Blockly.Luge.functionNames_ = Object.create(null);

  if (!Blockly.Luge.variableDB_) {
    Blockly.Luge.variableDB_ =
        new Blockly.Names(Blockly.Luge.RESERVED_WORDS_);
  } else {
    Blockly.Luge.variableDB_.reset();
  }

  Blockly.Luge.variableDB_.setVariableMap(workspace.getVariableMap());

  // Déclaration des variables

  var defvars = [];
  defvars.push("// TAG du fichier : " + Blockly.Luge.FileTAG);
  // defvars.push("");
  // defvars.push("// Declaration des variables");
  
  // Add user variables, but only ones that are being used.
  // var variables = Blockly.Variables.allUsedVarModels(workspace);
  
  // var all_variables = workspace.getAllVariables();
  
  // console.log("\nListe des variables");
  // for (var i = 0; i < variables.length; i++) {
	// var varName = Blockly.Luge.variableDB_.getName(variables[i].getId(), Blockly.Variables.NAME_TYPE);
	// var varType = Blockly.Variables.getVariable(workspace, variables[i].getId()).type;
	
	// var decalareType = "Booleen";
	// var initValue = "Booleen";
	
	// var recordVar = "";
	// var tagVar = "";
	
	// console.log(varName + " de type " + varType);
	
	// switch (varType) {
		// case 'String':
			// decalareType = 'Chaine';
			// initValue = '""';
			// break;
		// case 'Number':
			// decalareType = 'Numerique';
			// initValue = '0';
			// break;
		// case 'Boolean':
			// decalareType = 'Booleen';
			// initValue = 'Faux';
			// break;
		// default :
			// break;
	// }
	
    // defvars.push("Declare ( " + varName + ' , ' + decalareType + ' , ' + initValue + ' , ' + recordVar + ' )');
  // }

  // Blockly.Luge.definitions_['variables'] = defvars.join('\n');
  
  // Inclusion des scenarii

  // var includes = [];
  // includes.push("// Appel des scenarii externes");
 
  // for (var blockID in workspace.blockDB_) {
    // if (workspace.blockDB_[blockID].type == "include") {
      // includes.push('Inclure ( "' + workspace.blockDB_[blockID].getFieldValue('SCENARIO').replace('C:\\', 'c:\\') + '" )');
    // }
  // }  
    
  // includes.push("");

  // Blockly.Luge.definitions_['includes'] = includes.join('\n');
};

/**
 * Prepend the generated code with the variable definitions.
 * @param {string} code Generated code.
 * @return {string} Completed code.
 */
Blockly.Luge.finish = function(code) {
  // Convert the definitions dictionary into a list.
  var imports = [];
  var definitions = [];
  for (var name in Blockly.Luge.definitions_) {
    var def = Blockly.Luge.definitions_[name];
    definitions.push(def);
  }
  // Clean up temporary data.
  delete Blockly.Luge.definitions_;
  delete Blockly.Luge.functionNames_;
  Blockly.Luge.variableDB_.reset();
  var allDefs = imports.join('\n') + '\n\n' + definitions.join('\n\n');
  return allDefs.replace(/\n\n+/g, '\n\n').replace(/\n*$/, '\n\n\n') + code;
};

/**
 * Naked values are top-level blocks with outputs that aren't plugged into
 * anything.
 * @param {string} line Line of generated code.
 * @return {string} Legal line of code.
 */
Blockly.Luge.scrubNakedValue = function(line) {
  return line + '\n';
};

/**
 * Encode a string as a properly escaped Python string, complete with quotes.
 * @param {string} string Text to encode.
 * @return {string} Python string.
 * @private
 */
Blockly.Luge.quote_ = function(string) {
  // Can't use goog.string.quote since % must also be escaped.
  string = string.replace(/\\/g, '\\\\')
                 .replace(/\n/g, '\\\n');

  // Follow the CPython behaviour of repr() for a non-byte string.
  var quote = '\'';
  if (string.indexOf('\'') !== -1) {
    if (string.indexOf('"') === -1) {
      quote = '"';
    } else {
      string = string.replace(/'/g, '\\\'');
    }
  };
  return quote + string + quote;
};

/**
 * Common tasks for generating Python from blocks.
 * Handles comments for the specified block and any connected value blocks.
 * Calls any statements following this block.
 * @param {!Blockly.Block} block The current block.
 * @param {string} code The Python code created for this block.
 * @param {boolean=} opt_thisOnly True to generate code for only this statement.
 * @return {string} Python code with comments and subsequent blocks added.
 * @private
 */
Blockly.Luge.scrub_ = function(block, code, opt_thisOnly) {
  var commentCode = '';
  // Only collect comments for blocks that aren't inline.
  if (!block.outputConnection || !block.outputConnection.targetConnection) {
    // Collect comment for this block.
    var comment = block.getCommentText();
    comment = Blockly.utils.wrap(comment, Blockly.Luge.COMMENT_WRAP - 3);
    if (comment) {
      if (block.getProcedureDef) {
        // Use a comment block for function comments.
        commentCode += '"""' + comment + '\n"""\n';
      } else {
        commentCode += Blockly.Luge.prefixLines(comment + '\n', '# ');
      }
    }
    // Collect comments for all value arguments.
    // Don't collect comments for nested statements.
    for (var i = 0; i < block.inputList.length; i++) {
      if (block.inputList[i].type == Blockly.INPUT_VALUE) {
        var childBlock = block.inputList[i].connection.targetBlock();
        if (childBlock) {
          var comment = Blockly.Luge.allNestedComments(childBlock);
          if (comment) {
            commentCode += Blockly.Luge.prefixLines(comment, '# ');
          }
        }
      }
    }
  }
  var nextBlock = block.nextConnection && block.nextConnection.targetBlock();
  var nextCode = opt_thisOnly ? '' : Blockly.Luge.blockToCode(nextBlock);
  return commentCode + code + nextCode;
};

/**
 * Gets a property and adjusts the value, taking into account indexing, and
 * casts to an integer.
 * @param {!Blockly.Block} block The block.
 * @param {string} atId The property ID of the element to get.
 * @param {number=} opt_delta Value to add.
 * @param {boolean=} opt_negate Whether to negate the value.
 * @return {string|number}
 */
Blockly.Luge.getAdjustedInt = function(block, atId, opt_delta, opt_negate) {
  var delta = opt_delta || 0;
  if (block.workspace.options.oneBasedIndex) {
    delta--;
  }
  var defaultAtIndex = block.workspace.options.oneBasedIndex ? '1' : '0';
  var atOrder = delta ? Blockly.Luge.ORDER_ADDITIVE :
      Blockly.Luge.ORDER_NONE;
  var at = Blockly.Luge.valueToCode(block, atId, atOrder) || defaultAtIndex;

  if (Blockly.isNumber(at)) {
    // If the index is a naked number, adjust it right now.
    at = parseInt(at, 10) + delta;
    if (opt_negate) {
      at = -at;
    }
  } else {
    // If the index is dynamic, adjust it in code.
    if (delta > 0) {
      at = 'int(' + at + ' + ' + delta + ')';
    } else if (delta < 0) {
      at = 'int(' + at + ' - ' + -delta + ')';
    } else {
      at = 'int(' + at + ')';
    }
    if (opt_negate) {
      at = '-' + at;
    }
  }
  return at;
};
