/**
 * @license
 * Visual Blocks Language
 *
 * Copyright 2012 Google Inc.
 * https://developers.google.com/blockly/
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *   http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/**
 * @fileoverview Generating Python for logic blocks.
 * @author q.neutron@gmail.com (Quynh Neutron)
 */
'use strict';

goog.provide('Blockly.Luge.logic');

goog.require('Blockly.Luge');


Blockly.Luge['controls_if'] = function(block) {
  // If/elseif/else condition.
  var n = 0;
  var code = '', branchCode, conditionCode;
  do {
    conditionCode = Blockly.Luge.valueToCode(block, 'IF' + n,
        Blockly.Luge.ORDER_NONE) || 'False';
    branchCode = Blockly.Luge.statementToCode(block, 'DO' + n) ||
        Blockly.Luge.PASS;
    code += (n == 0 ? 'if ' : 'elif ' ) + conditionCode + ':\n' + branchCode;

    ++n;
  } while (block.getInput('IF' + n));

  if (block.getInput('ELSE')) {
    branchCode = Blockly.Luge.statementToCode(block, 'ELSE') ||
        Blockly.Luge.PASS;
    code += 'else:\n' + branchCode;
  }
  return code;
};

Blockly.Luge['controls_ifelse'] = Blockly.Luge['controls_if'];

Blockly.Luge['logic_compare'] = function(block) {
  // Comparison operator.
  var OPERATORS = {
    'EQ': '==',
    'NEQ': '!=',
    'LT': '<',
    'LTE': '<=',
    'GT': '>',
    'GTE': '>='
  };
  var operator = OPERATORS[block.getFieldValue('OP')];
  var order = Blockly.Luge.ORDER_RELATIONAL;
  var argument0 = Blockly.Luge.valueToCode(block, 'A', order) || '0';
  var argument1 = Blockly.Luge.valueToCode(block, 'B', order) || '0';
  var code = argument0 + ' ' + operator + ' ' + argument1;
  return [code, order];
};

Blockly.Luge['logic_operation'] = function(block) {
  // Operations 'and', 'or'.
  var operator = (block.getFieldValue('OP') == 'AND') ? 'and' : 'or';
  var order = (operator == 'and') ? Blockly.Luge.ORDER_LOGICAL_AND :
      Blockly.Luge.ORDER_LOGICAL_OR;
  var argument0 = Blockly.Luge.valueToCode(block, 'A', order);
  var argument1 = Blockly.Luge.valueToCode(block, 'B', order);
  if (!argument0 && !argument1) {
    // If there are no arguments, then the return value is false.
    argument0 = 'False';
    argument1 = 'False';
  } else {
    // Single missing arguments have no effect on the return value.
    var defaultArgument = (operator == 'and') ? 'True' : 'False';
    if (!argument0) {
      argument0 = defaultArgument;
    }
    if (!argument1) {
      argument1 = defaultArgument;
    }
  }
  var code = argument0 + ' ' + operator + ' ' + argument1;
  return [code, order];
};

Blockly.Luge['logic_negate'] = function(block) {
  // Negation.
  var argument0 = Blockly.Luge.valueToCode(block, 'BOOL',
      Blockly.Luge.ORDER_LOGICAL_NOT) || 'True';
  var code = 'not ' + argument0;
  return [code, Blockly.Luge.ORDER_LOGICAL_NOT];
};

Blockly.Luge['logic_boolean'] = function(block) {
  // Boolean values true and false.
  var code = (block.getFieldValue('BOOL') == 'TRUE') ? 'True' : 'False';
  return [code, Blockly.Luge.ORDER_ATOMIC];
};

Blockly.Luge['logic_null'] = function(block) {
  // Null data type.
  return ['None', Blockly.Luge.ORDER_ATOMIC];
};

Blockly.Luge['logic_ternary'] = function(block) {
  // Ternary operator.
  var value_if = Blockly.Luge.valueToCode(block, 'IF',
      Blockly.Luge.ORDER_CONDITIONAL) || 'False';
  var value_then = Blockly.Luge.valueToCode(block, 'THEN',
      Blockly.Luge.ORDER_CONDITIONAL) || 'None';
  var value_else = Blockly.Luge.valueToCode(block, 'ELSE',
      Blockly.Luge.ORDER_CONDITIONAL) || 'None';
  var code = value_then + ' if ' + value_if + ' else ' + value_else;
  return [code, Blockly.Luge.ORDER_CONDITIONAL];
};
