/**
 * Blockly Demos: Code
 *
 * Copyright 2012 Google Inc.
 * https://developers.google.com/blockly/
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *   http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/**
 * @fileoverview JavaScript for Blockly's Code demo.
 * @author fraser@google.com (Neil Fraser)
 */
'use strict';

/**
 * Create a namespace for the application.
 */
var Code = {};

/**
 * Lookup for names of supported languages.  Keys should be in ISO 639 format.
 */
Code.LANGUAGE_NAME = {
  'en': 'English',
  'fr': 'Français'
};

/**
 * List of RTL languages.
 */
Code.LANGUAGE_RTL = ['ar', 'fa', 'he', 'lki'];

/**
 * Blockly's main workspace.
 * @type {Blockly.WorkspaceSvg}
 */
Code.workspace = null;

/**
 * Extracts a parameter from the URL.
 * If the parameter is absent default_value is returned.
 * @param {string} name The name of the parameter.
 * @param {string} defaultValue Value to return if parameter not found.
 * @return {string} The parameter value or the default value if not found.
 */
Code.getStringParamFromUrl = function(name, defaultValue) {
  var val = location.search.match(new RegExp('[?&]' + name + '=([^&]+)'));
  return val ? decodeURIComponent(val[1].replace(/\+/g, '%20')) : defaultValue;
};


/**
 * Create variable buttons callbacks.
 */
 
Code.createBoolVariableCallback = function (button) {
	alert("bool");
	Blockly.Variables.createVariable(button.getTargetWorkspace(), null, 'Boolean');
} 

Code.createStringVariableCallback = function (button) {
	alert("string");
	Blockly.Variables.createVariable(button.getTargetWorkspace(), null, 'String');
} 

Code.createNumVariableCallback = function (button) {
	alert("num");
	Blockly.Variables.createVariable(button.getTargetWorkspace(), null, 'Number');
}

/**
 * Get the language of this user from the URL.
 * @return {string} User's language.
 */
Code.getLang = function() {
  var lang = Code.getStringParamFromUrl('lang', '');
  if (Code.LANGUAGE_NAME[lang] === undefined) {
    // Default to English.
    lang = 'fr';
  }
  return lang;
};

/**
 * Is the current language (Code.LANG) an RTL language?
 * @return {boolean} True if RTL, false if LTR.
 */
Code.isRtl = function() {
  return Code.LANGUAGE_RTL.indexOf(Code.LANG) != -1;
};

/**
 * Load blocks saved on App Engine Storage or in session/local storage.
 * @param {string} defaultXml Text representation of default blocks.
 */
Code.loadBlocks = function(defaultXml) {
  try {
    var loadOnce = window.sessionStorage.loadOnceBlocks;
  } catch(e) {
    // Firefox sometimes throws a SecurityError when accessing sessionStorage.
    // Restarting Firefox fixes this, so it looks like a bug.
    var loadOnce = null;
  }
  if ('BlocklyStorage' in window && window.location.hash.length > 1) {
    // An href with #key trigers an AJAX call to retrieve saved blocks.
    BlocklyStorage.retrieveXml(window.location.hash.substring(1));
  } else if (loadOnce) {
    // Language switching stores the blocks during the reload.
    delete window.sessionStorage.loadOnceBlocks;
    var xml = Blockly.Xml.textToDom(loadOnce);
    Blockly.Xml.domToWorkspace(xml, Code.workspace);
  } else if (defaultXml) {
    // Load the editor with default starting blocks.
    var xml = Blockly.Xml.textToDom(defaultXml);
    Blockly.Xml.domToWorkspace(xml, Code.workspace);
  } else if ('BlocklyStorage' in window) {
    // Restore saved blocks in a separate thread so that subsequent
    // initialization is not affected from a failed load.
    window.setTimeout(BlocklyStorage.restoreBlocks, 0);
  }
};

/**
 * Save the blocks and reload with a different language.
 */
Code.changeLanguage = function() {
  // Store the blocks for the duration of the reload.
  // MSIE 11 does not support sessionStorage on file:// URLs.
  if (window.sessionStorage) {
    var xml = Blockly.Xml.workspaceToDom(Code.workspace);
    var text = Blockly.Xml.domToText(xml);
    window.sessionStorage.loadOnceBlocks = text;
  }

  var languageMenu = document.getElementById('languageMenu');
  var newLang = encodeURIComponent(
      languageMenu.options[languageMenu.selectedIndex].value);
  var search = window.location.search;
  if (search.length <= 1) {
    search = '?lang=' + newLang;
  } else if (search.match(/[?&]lang=[^&]*/)) {
    search = search.replace(/([?&]lang=)[^&]*/, '$1' + newLang);
  } else {
    search = search.replace(/\?/, '?lang=' + newLang + '&');
  }

  window.location = window.location.protocol + '//' +
      window.location.host + window.location.pathname + search;
};


/*
 * Makes the currently selected block a user-generated shadow block. These
 * blocks are not made into real shadow blocks, but recorded in the model
 * and visually marked as shadow blocks, allowing the user to move and edit
 * them (which would be impossible with actual shadow blocks). Updates the
 * preview when done.
 */
Code.addShadow = function() {
  // No block selected to make a shadow block.
  if (!Blockly.selected) {
    return;
  }
  // Clear any previous warnings on the block (would only have warnings on
  // a non-shadow block if it was nested inside another shadow block).
  Blockly.selected.setWarningText(null);
  // Set selected block and all children as shadow blocks.
  this.addShadowForBlockAndChildren_(Blockly.selected);

  // Save and update the preview.
  this.saveStateFromWorkspace();
  this.updatePreview();
};

/**
 * Sets a block and all of its children to be user-generated shadow blocks,
 * both in the model and view.
 * @param {!Blockly.Block} block The block to be converted to a user-generated
 *    shadow block.
 * @private
 */
Code.addShadowForBlockAndChildren_ =
    function(block) {
  // Convert to shadow block.
  this.view.markShadowBlock(block);
  this.model.addShadowBlock(block.id);

  if (FactoryUtils.hasVariableField(block)) {
    block.setWarningText('Cannot make variable blocks shadow blocks.');
  }

  // Convert all children to shadow blocks recursively.
  var children = block.getChildren();
  for (var i = 0; i < children.length; i++) {
    this.addShadowForBlockAndChildren_(children[i]);
  }
};

/**
 * If the currently selected block is a user-generated shadow block, this
 * function makes it a normal block again, removing it from the list of
 * shadow blocks and loading the workspace again. Updates the preview again.
 */
Code.removeShadow = function() {
  // No block selected to modify.
  if (!Blockly.selected) {
    return;
  }
  this.model.removeShadowBlock(Blockly.selected.id);
  this.view.unmarkShadowBlock(Blockly.selected);

  // If turning invalid shadow block back to normal block, remove warning.
  Blockly.selected.setWarningText(null);

  this.saveStateFromWorkspace();
  this.updatePreview();
};

/**
 * Given a unique block ID, uses the model to determine if a block is a
 * user-generated shadow block.
 * @param {string} blockId The unique ID of the block to examine.
 * @return {boolean} True if the block is a user-generated shadow block, false
 *    otherwise.
 */
Code.isUserGenShadowBlock = function(blockId) {
  return this.model.isShadowBlock(blockId);
};

/**
 * Call when importing XML containing real shadow blocks. This function turns
 * all real shadow blocks loaded in the workspace into user-generated shadow
 * blocks, meaning they are marked as shadow blocks by the model and appear as
 * shadow blocks in the view but are still editable and movable.
 */
Code.convertShadowBlocks = function() {
  var blocks = Code.workspace.getAllBlocks(false);
  for (var i = 0, block; block = blocks[i]; i++) {
    if (block.isShadow()) {
      block.setShadow(false);
      // Delete the shadow DOM attached to the block so that the shadow block
      // does not respawn. Dependent on implementation details.
      var parentConnection = block.outputConnection ?
          block.outputConnection.targetConnection :
          block.previousConnection.targetConnection;
      if (parentConnection) {
        parentConnection.setShadowDom(null);
      }
      //this.model.addShadowBlock(block.id);
      //this.view.markShadowBlock(block);
    }
  }
};

/**
 * Bind a function to a button's click event.
 * On touch enabled browsers, ontouchend is treated as equivalent to onclick.
 * @param {!Element|string} el Button element or ID thereof.
 * @param {!Function} func Event handler to bind.
 */
Code.bindClick = function(el, func) {
  if (typeof el == 'string') {
    el = document.getElementById(el);
  }
  el.addEventListener('click', func, true);
  el.addEventListener('touchend', func, true);
};

Code.onPlaceBlock = function onPlaceBlock(event) {
  //console.log("Event type: " + event.type);
  if (event.type == Blockly.Events.BLOCK_MOVE ) {
	  // fix shadow blocks
	  Code.workspace.render();
  }
  
  else if (event.type == Blockly.Events.BLOCK_CREATE ) {
	var block = Blockly.mainWorkspace.getBlockById(event.blockId);
	//console.log("Block created");
    block.setCollapsed(false);
	// convert shadow blocks into real blocks
	Code.convertShadowBlocks();
	
	var index_loop = -1;
	
	for (var ID in Blockly.mainWorkspace.blockDB_) {
		var temp_block = Blockly.mainWorkspace.blockDB_[ID];
		if (((temp_block.type == "for_loop")||(temp_block.type == "bky_for_loop")||(temp_block.type == "bky_while_loop"))&&(ID != event.blockId)) {
			var temp_array = temp_block.getFieldValue("LOOP_INDEX").split("_");	
			if (temp_array[temp_array.length-1].trim().length > 0) {					
				if (parseInt(temp_array[temp_array.length-1])>index_loop) {
					index_loop = parseInt(temp_array[temp_array.length-1]);		
				}
			}
		}
	  }
	  
	  index_loop += 1;
	
	var block_type = block.type;
    if ((block_type == "for_loop")||(block_type == "bky_for_loop")||(block_type == "bky_while_loop")) {
		block.setFieldValue("ForLoop_i_"+index_loop.toString().padStart(2, '0') , "LOOP_INDEX");
		var myNewVariable = Blockly.mainWorkspace.createVariable("ForLoop_i_"+index_loop.toString().padStart(2, '0'), 'Number');
		var childBlock = Blockly.mainWorkspace.newBlock('variables_get_dynamic_num');
		childBlock.setFieldValue(myNewVariable.getId(), 'VAR');
		childBlock.setDeletable(false);
		childBlock.setEditable(false);
		childBlock.setMovable(false);
		childBlock.initSvg();
		childBlock.render();

		var parentConnection = block.getInput('I_LOOP').connection;
		var childConnection = childBlock.outputConnection;
		parentConnection.connect(childConnection);
		
		if (block_type != "bky_while_loop") {
			myNewVariable = Blockly.mainWorkspace.createVariable("ForLoop_i_"+index_loop.toString().padStart(2, '0') + "_max", 'Number');
			childBlock = Blockly.mainWorkspace.newBlock('variables_get_dynamic_num');
			childBlock.setFieldValue(myNewVariable.getId(), 'VAR');
			childBlock.setDeletable(false);
			childBlock.setEditable(false);
			childBlock.setMovable(false);
			childBlock.initSvg();
			childBlock.render();

			parentConnection = block.getInput('I_LOOP_MAX').connection;
			childConnection = childBlock.outputConnection;
			parentConnection.connect(childConnection);
		}
		index_loop++;
    }
  }
}

Code.onDeleteBlock = function onPlaceBlock(event) {
  if (event.type == Blockly.Events.BLOCK_DELETE ) {	
	  var for_loop_list_var = [];
	  
	  for (var blockID in Blockly.mainWorkspace.blockDB_) {
		var block_type = Blockly.mainWorkspace.blockDB_[blockID].type;
		if ((block_type == "for_loop")||(block_type == "bky_for_loop")||(block_type == "bky_while_loop")) {
			for_loop_list_var.push(Blockly.mainWorkspace.blockDB_[blockID].getFieldValue("LOOP_INDEX"));
			if(block_type != "bky_while_loop") {
				for_loop_list_var.push(Blockly.mainWorkspace.blockDB_[blockID].getFieldValue("LOOP_INDEX") + "_max");
			}
		}
	  }
	  
	  var all_variables = Blockly.mainWorkspace.getAllVariables();
	  
	  for (var i = 0; i<all_variables.length; i++) {
		  if ((for_loop_list_var.indexOf(all_variables[i].name) < 0)&&(all_variables[i].name.startsWith("for_loop_i"))) Blockly.mainWorkspace.deleteVariableById(all_variables[i].getId());
	  }
  }
}

Code.onVarRename = function onVarRename(event) {
	var block = Blockly.mainWorkspace.getBlockById(event.blockId);
	if (block) {
		var typeBlock = block.type;
		var newVarId = event.newValue;
		var newType = "";
		
		if ((event.type == Blockly.Events.BLOCK_CHANGE ) && (event.element == "field" )) {	
			if (typeBlock.startsWith("variables_get_")) {
				switch (Blockly.Variables.getVariable(Code.workspace, newVarId).type) {
					case 'String':
						block.type = 'variables_get_dynamic_string';
						block.setColour("%{BKY_TEXTS_HUE}");
						break;
					case 'Number':
						block.type = 'variables_get_dynamic_num';
						block.setColour("%{BKY_MATH_HUE}");
						break;
					case 'Boolean':
						block.type = 'variables_get_dynamic_bool';
						block.setColour("%{BKY_LOGIC_HUE}");
						break;
					default :
						break;
				}		
			   block.outputConnection.setCheck(Blockly.Variables.getVariable(Code.workspace, newVarId).type);
			}
			else if (typeBlock.startsWith("variables_set_")) {
				switch (Blockly.Variables.getVariable(Code.workspace, newVarId).type) {
					case 'String':
						block.type = 'variables_set_dynamic_string';
						block.setColour(165);
						break;
					case 'Number':
						block.type = 'variables_set_dynamic_num';
						block.setColour(240);
						break;
					case 'Boolean':
						block.type = 'variables_set_dynamic_bool';
						block.setColour(120);
						break;
					default :
						break;
				}		
			   block.getInput("VALUE").setCheck(Blockly.Variables.getVariable(Code.workspace, newVarId).type);
			}
		}
	}
}

Code.onChangeBlock = function onChangeBlock(event) {
  	var xhr = new XMLHttpRequest();
	xhr.open("POST", "http://127.0.0.1:".concat("8888", "/SetXml"), true);

	//Send the proper header information along with the request
	xhr.setRequestHeader("Content-Type", "application/x-www-form-urlencoded");

	xhr.onreadystatechange = function() { // Call a function when the state changes.
		if (this.readyState === XMLHttpRequest.DONE && this.status === 200) {
			// Request finished. Do processing here.
		}
	}
	xhr.send(Blockly.Xml.domToPrettyText(Blockly.Xml.workspaceToDom(Blockly.mainWorkspace, false)));
}

/**
 * Load the Prettify CSS and JavaScript.
 */
Code.importPrettify = function() {
  var script = document.createElement('script');
  script.setAttribute('src', 'https://cdn.rawgit.com/google/code-prettify/master/loader/run_prettify.js');
  document.head.appendChild(script);
};

/**
 * Compute the absolute coordinates and dimensions of an HTML element.
 * @param {!Element} element Element to match.
 * @return {!Object} Contains height, width, x, and y properties.
 * @private
 */
Code.getBBox_ = function(element) {
  var height = element.offsetHeight;
  var width = element.offsetWidth;
  var x = 0;
  var y = 0;
  do {
    x += element.offsetLeft;
    y += element.offsetTop;
    element = element.offsetParent;
  } while (element);
  return {
    height: height,
    width: width,
    x: x,
    y: y
  };
};

/**
 * User's language (e.g. "en").
 * @type {string}
 */
Code.LANG = Code.getLang();

/**
 * List of tab names.
 * @private
 */
Code.TABS_ = ['blocks', 'xml', 'LUGE'];

Code.selected = 'blocks';

/**
 * Switch the visible pane when a tab is clicked.
 * @param {string} clickedName Name of tab clicked.
 */
Code.tabClick = function(clickedName) {
  // If the XML tab was open, save and render the content.
  if (document.getElementById('tab_xml').className == 'tabon') {
    var xmlTextarea = document.getElementById('content_xml');
    var xmlText = xmlTextarea.value;
    var xmlDom = null;
    try {
      xmlDom = Blockly.Xml.textToDom(xmlText);
    } catch (e) {
      var q =
          window.confirm(MSG['badXml'].replace('%1', e));
      if (!q) {
        // Leave the user on the XML tab.
        return;
      }
    }
    if (xmlDom) {
      Code.workspace.clear();
      Blockly.Xml.domToWorkspace(xmlDom, Code.workspace);
    }
  }

  if (document.getElementById('tab_blocks').className == 'tabon') {
    Code.workspace.setVisible(false);
  }
  // Deselect all tabs and hide all panes.
  for (var i = 0; i < Code.TABS_.length; i++) {
    var name = Code.TABS_[i];
    document.getElementById('tab_' + name).className = 'taboff';
    document.getElementById('content_' + name).style.visibility = 'hidden';
  }

  // Select the active tab.
  Code.selected = clickedName;
  document.getElementById('tab_' + clickedName).className = 'tabon';
  // Show the selected pane.
  document.getElementById('content_' + clickedName).style.visibility =
      'visible';
  Code.renderContent();
  if (clickedName == 'blocks') {
    Code.workspace.setVisible(true);
  }
  Blockly.svgResize(Code.workspace);
};

/**
 * Populate the currently selected pane with content generated from the blocks.
 */
Code.renderContent = function() {
  var content = document.getElementById('content_' + Code.selected);
  // Initialize the pane.
  if (content.id == 'content_xml') {
    var xmlTextarea = document.getElementById('content_xml');
    var xmlDom = Blockly.Xml.workspaceToDom(Code.workspace);
    var xmlText = Blockly.Xml.domToPrettyText(xmlDom);
    xmlTextarea.value = xmlText;
    xmlTextarea.focus();
  } else if (content.id == 'content_LUGE') {
    Code.attemptCodeGeneration(Blockly.Luge, 'c');	//TODO : trouver la bonne coloration syntaxique
  }
};

/**
 * Attempt to generate the code and display it in the UI, pretty printed.
 * @param generator {!Blockly.Generator} The generator to use.
 * @param prettyPrintType {string} The file type key for the pretty printer.
 */
Code.attemptCodeGeneration = function(generator, prettyPrintType) {
  var content = document.getElementById('content_' + Code.selected);
  content.textContent = '';
  if (Code.checkAllGeneratorFunctionsDefined(generator)) {
    var code = generator.workspaceToCode(Code.workspace);

    content.textContent = code;
    if (typeof PR.prettyPrintOne == 'function') {
      code = content.textContent;
      code = PR.prettyPrintOne(code, prettyPrintType);
      content.innerHTML = code;
    }
  }
};

/**
 * Check whether all blocks in use have generator functions.
 * @param generator {!Blockly.Generator} The generator to use.
 */
Code.checkAllGeneratorFunctionsDefined = function(generator) {
  var blocks = Code.workspace.getAllBlocks(false);
  var missingBlockGenerators = [];
  for (var i = 0; i < blocks.length; i++) {
    var blockType = blocks[i].type;
    if (!generator[blockType]) {
      if (missingBlockGenerators.indexOf(blockType) === -1) {
        missingBlockGenerators.push(blockType);
      }
    }
  }

  var valid = missingBlockGenerators.length == 0;
  if (!valid) {
    var msg = 'The generator code for the following blocks not specified for '
        + generator.name_ + ':\n - ' + missingBlockGenerators.join('\n - ');
    Blockly.alert(msg);  // Assuming synchronous. No callback.
  }
  return valid;
};

/**
 * Initialize Blockly.  Called on page load.
 */
Code.init = function() {
  Code.initLanguage();

  var rtl = Code.isRtl();
  var container = document.getElementById('content_area');
  var onresize = function(e) {
    var bBox = Code.getBBox_(container);
    for (var i = 0; i < Code.TABS_.length; i++) {
      var el = document.getElementById('content_' + Code.TABS_[i]);
      el.style.top = bBox.y + 'px';
      el.style.left = bBox.x + 'px';
      // Height and width need to be set, read back, then set again to
      // compensate for scrollbars.
      el.style.height = bBox.height + 'px';
      el.style.height = (2 * bBox.height - el.offsetHeight) + 'px';
      el.style.width = bBox.width + 'px';
      el.style.width = (2 * bBox.width - el.offsetWidth) + 'px';
    }
    // Make the 'Blocks' tab line up with the toolbox.
    if (Code.workspace) {
      document.getElementById('tab_blocks').style.minWidth =
          (38) + 'px';
          // Account for the 19 pixel margin and on each side.
    }
  };
  window.addEventListener('resize', onresize, false);

  // The toolbox XML specifies each category name using Blockly's messaging
  // format (eg. `<category name="%{BKY_CATLOGIC}">`).
  // These message keys need to be defined in `Blockly.Msg` in order to
  // be decoded by the library. Therefore, we'll use the `MSG` dictionary that's
  // been defined for each language to import each category name message
  // into `Blockly.Msg`.
  // TODO: Clean up the message files so this is done explicitly instead of
  // through this for-loop.
  for (var messageKey in MSG) {
    if (messageKey.indexOf('cat') == 0) {
      Blockly.Msg[messageKey.toUpperCase()] = MSG[messageKey];
    }
  }

  // Construct the toolbox XML, replacing translated variable names.
  var toolboxText = document.getElementById('toolbox').outerHTML;
  toolboxText = toolboxText.replace(/(^|[^%]){(\w+)}/g,
      function(m, p1, p2) {return p1 + MSG[p2];});
  var toolboxXml = Blockly.Xml.textToDom(toolboxText);

  Code.workspace = Blockly.inject('content_blocks',
      {grid:
          {spacing: 25,
           length: 3,
           colour: '#ccc',
           snap: true},
       readOnly: true,
       media: '../google-blockly/media/',
       rtl: rtl,
       toolbox: toolboxXml,
       zoom:
           {controls: true,
            wheel: true},
       trashcan: false
      });

  // Add to reserved word list: Local variables in execution environment (runJS)
  // and the infinite loop detection function.
  // Blockly.JavaScript.addReservedWords('code,timeouts,checkTimeout');

  Code.loadBlocks('');

  if ('BlocklyStorage' in window) {
    // Hook a save function onto unload.
    BlocklyStorage.backupOnUnload(Code.workspace);
  }

  Code.tabClick(Code.selected);

  Code.bindClick('trashButton',
      function() {Code.discard(); Code.renderContent();});
  // Code.bindClick('runButton', Code.runJS);
  // Disable the link button if page isn't backed by App Engine storage.
  // var linkButton = document.getElementById('linkButton');
  // if ('BlocklyStorage' in window) {
    // BlocklyStorage['HTTPREQUEST_ERROR'] = MSG['httpRequestError'];
    // BlocklyStorage['LINK_ALERT'] = MSG['linkAlert'];
    // BlocklyStorage['HASH_ERROR'] = MSG['hashError'];
    // BlocklyStorage['XML_ERROR'] = MSG['xmlError'];
    // Code.bindClick(linkButton,
        // function() {BlocklyStorage.link(Code.workspace);});
  // } else if (linkButton) {
    // linkButton.className = 'disabled';
  // }

  for (var i = 0; i < Code.TABS_.length; i++) {
    var name = Code.TABS_[i];
    Code.bindClick('tab_' + name,
        function(name_) {return function() {Code.tabClick(name_);};}(name));
  }
  onresize();
  Blockly.svgResize(Code.workspace);
  Code.workspace.addChangeListener(Code.onPlaceBlock);
  Code.workspace.addChangeListener(Code.onDeleteBlock);
  Code.workspace.addChangeListener(Code.onChangeBlock);
  Code.workspace.addChangeListener(Code.onVarRename);
  // Code.workspace.addChangeListener(Code.onNewParent);
  Code.workspace.addChangeListener(Blockly.Events.disableOrphans);  
  
  //Code.workspace.registerToolboxCategoryCallback('VASCO_VARIABLE', Blockly.Luge.variablesFlyoutCallback);


  // Lazy-load the syntax-highlighting.
  window.setTimeout(Code.importPrettify, 1);
};

/**
 * Initialize the page language.
 */
Code.initLanguage = function() {
  // Set the HTML's language and direction.
  var rtl = Code.isRtl();
  document.dir = rtl ? 'rtl' : 'ltr';
  document.head.parentElement.setAttribute('lang', Code.LANG);

  // Sort languages alphabetically.
  var languages = [];
  for (var lang in Code.LANGUAGE_NAME) {
    languages.push([Code.LANGUAGE_NAME[lang], lang]);
  }
  var comp = function(a, b) {
    // Sort based on first argument ('English', 'Русский', '简体字', etc).
    if (a[0] > b[0]) return 1;
    if (a[0] < b[0]) return -1;
    return 0;
  };
  languages.sort(comp);
  // Populate the language selection menu.
  // var languageMenu = document.getElementById('languageMenu');
  // languageMenu.options.length = 0;
  // for (var i = 0; i < languages.length; i++) {
    // var tuple = languages[i];
    // var lang = tuple[tuple.length - 1];
    // var option = new Option(tuple[0], lang);
    // if (lang == Code.LANG) {
      // option.selected = true;
    // }
    // languageMenu.options.add(option);
  // }
  // languageMenu.addEventListener('change', Code.changeLanguage, true);

  // Inject language strings.
  // document.title += ' ' + MSG['title'];
  // document.getElementById('title').textContent = MSG['title'];
  document.getElementById('tab_blocks').textContent = MSG['blocks'];

  // document.getElementById('linkButton').title = MSG['linkTooltip'];
  // document.getElementById('runButton').title = MSG['runTooltip'];
  document.getElementById('trashButton').title = MSG['trashTooltip'];
};

/**
 * Execute the user's code.
 * Just a quick and dirty eval.  Catch infinite loops.
 */
Code.runJS = function() {
  Blockly.JavaScript.INFINITE_LOOP_TRAP = '  checkTimeout();\n';
  var timeouts = 0;
  var checkTimeout = function() {
    if (timeouts++ > 1000000) {
      throw MSG['timeout'];
    }
  };
  var code = Blockly.JavaScript.workspaceToCode(Code.workspace);
  Blockly.JavaScript.INFINITE_LOOP_TRAP = null;
  try {
    eval(code);
  } catch (e) {
    alert(MSG['badCode'].replace('%1', e));
  }
};

/**
 * Discard all blocks from the workspace.
 */
Code.discard = function() {
  var count = Code.workspace.getAllBlocks(false).length;
  if (count < 2 ||
      window.confirm(Blockly.Msg['DELETE_ALL_BLOCKS'].replace('%1', count))) {
    Code.workspace.clear();
    if (window.location.hash) {
      window.location.hash = '';
    }
  }
};


/* Override blockly core */
Blockly.BlockSvg.prototype.setHighlighted = function(highlighted) {
  if (!this.rendered) {
    return;
  }
  if (highlighted) {
    this.addSelect();
    
    this.svgPath_.setAttribute('filter',
        'url(#' + this.workspace.options.embossFilterId + ')');
    this.svgPathLight_.style.display = 'none';
    
  } else {
    this.removeSelect();
    
    Blockly.utils.removeAttribute(this.svgPath_, 'filter');
    delete this.svgPathLight_.style.display;
    
  }
};

var scripts= document.getElementsByTagName('script');
var path= scripts[scripts.length-1].src.split('?')[0];      // remove any ?query
var mydir= path.split('/').slice(0, -1).join('/')+'/';  // remove last filename part of path

// alert(mydir);

// Load the Code demo's language strings.
document.write('<script src="' + mydir + 'msg/' + Code.LANG + '.js"></script>\n');
// Load Blockly's language strings.
document.write('<script src="' + mydir + '../google-blockly/msg/js/' + Code.LANG + '.js"></script>\n');

window.addEventListener('load', Code.init);
